// HowToReadAndWriteNex5Files.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "Nex5File.h"
#include <vector>
#include <algorithm>

// To make code cleaner, I did not use explicit conversions from __int64 to long, etc. 
// Let's suppress conversion warning (do not do this in your own code!)
#pragma warning( disable : 4244)
// Let's also suppress Microsoft insecure function warning (do not do this in your own code!)
#pragma warning( disable : 4996)


#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif


void ReadAllFromNex5File( const char* filePath );
void PrintHeaderInfo( Nex5FileHeader &fh );
std::vector<double> ReadTimestampsReturnTimestampsInSeconds( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader );
std::vector<double> ReadWaveformOrContValuesReturnValuesInMilliVolts( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader );
void ReadNeuronData( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader );
void ReadEventData( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader );
void ReadIntervalData( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader );
void ReadWaveformData( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader );
void ReadContinuousData( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader );
void ReadMarkerData( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader );
void WriteNex5File( const char* fname );


int main( int argc, char* argv[] )
{
    ReadAllFromNex5File( "TestDataFileForHowTo.nex5" );

    WriteNex5File( "TestFileWrittenByHowToProgram.nex5" );

    return 0;
}


// the code to read all data types from .nex5 file
void ReadAllFromNex5File( const char* filePath )
{
    Nex5FileHeader fileHeader;
    std::vector<Nex5VarHeader> varHeaders;

    FILE* fp = fopen( filePath, "rb" );

    if ( fp == NULL ) {
        printf( "unable to open file '%s'\n", filePath );
        return;
    }

    // 1. read file header
    fread( &fileHeader, sizeof( Nex5FileHeader ), 1, fp );
    PrintHeaderInfo( fileHeader );

    // 2. read the variable headers
    varHeaders.resize( fileHeader.NumberOfVariables );
    fread( &varHeaders[0], sizeof( Nex5VarHeader )*fileHeader.NumberOfVariables, 1, fp );

    // 3. read variable data
    for ( size_t i = 0; i < varHeaders.size(); ++i ) {
        switch ( varHeaders[i].Type ) {
            case NEX5_VARIABLE_TYPE_NEURON:
                ReadNeuronData( fp, varHeaders[i], fileHeader );
                break;
            case NEX5_VARIABLE_TYPE_EVENT:
                ReadEventData( fp, varHeaders[i], fileHeader );
                break;
            case NEX5_VARIABLE_TYPE_INTERVAL:
                ReadIntervalData( fp, varHeaders[i], fileHeader );
                break;
            case NEX5_VARIABLE_TYPE_WAVEFORM:
                ReadWaveformData( fp, varHeaders[i], fileHeader );
                break;
            case NEX5_VARIABLE_TYPE_POPULATION_VECTOR:
                printf( "skipping population vector...\n" );
                break;
            case NEX5_VARIABLE_TYPE_CONTINUOUS:
                ReadContinuousData( fp, varHeaders[i], fileHeader );
                break;
            case NEX5_VARIABLE_TYPE_MARKER:
                ReadMarkerData( fp, varHeaders[i], fileHeader );
                break;
            default:
                printf( "Invalid variable type %d\n", varHeaders[i].Type );
                break;
        }
    }

    // 4. read metadata
    if ( fileHeader.MetadataOffset > 0 ) {
        // find the length of the file
        fseek( fp, 0 , SEEK_END );
        unsigned int fileLength = ftell( fp );
        // check that metadata length is more than zero
        if ( fileHeader.MetadataOffset < fileLength ) {
            size_t metaLength = fileLength - fileHeader.MetadataOffset;
            std::vector<char> buf( metaLength );
            // seek to metadata start
            fseek( fp, fileHeader.MetadataOffset, SEEK_SET );
            // read metadata
            fread( &buf[0], metaLength, 1, fp );
            // add zero to terminate string
            buf.push_back( '\0' );
            printf( "metadata: %s\n", ( const char* )&buf[0] );
        }
    }

    fclose( fp );
}

std::vector<double> ReadTimestampsReturnTimestampsInSeconds( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader )
{
    std::vector<double> timestampsInSeconds;

    if ( varHeader.TimestampDataType == 0 ) {
        // timestamps are saved as 32-bit integers
        std::vector<int> timestamps;
        timestamps.resize( varHeader.Count );
        // read the timestamps, 4 bytes per timestamp
        fread( &timestamps[0], varHeader.Count * sizeof( int ), 1, fp );
        // convert timestamps to seconds
        for ( size_t i = 0; i < timestamps.size(); ++i ) {
            timestampsInSeconds.push_back( ( double )timestamps[i] / fileHeader.TimestampFrequency );
        }
    } else {
        // timestamps are saved as 64-bit integers
        std::vector<long long> timestamps;
        timestamps.resize( varHeader.Count );
        // read the timestamps, 8 bytes per timestamp
        fread( &timestamps[0], varHeader.Count * sizeof( long long ), 1, fp );
        // convert timestamps to seconds
        for ( size_t i = 0; i < timestamps.size(); ++i ) {
            timestampsInSeconds.push_back( ( double )timestamps[i] / fileHeader.TimestampFrequency );
        }
    }
    return timestampsInSeconds;
}

std::vector<double> ReadWaveformOrContValuesReturnValuesInMilliVolts( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader )
{
    std::vector<double> valuesInMilliVolts;

    unsigned long long totalNumberOfPoints = varHeader.NumberOfDataPoints;
    if ( varHeader.Type == NEX5_VARIABLE_TYPE_WAVEFORM ) {
        // for waveform variable, varHeader.NumberOfDataPoints is the number of data points in each waveform
        totalNumberOfPoints = varHeader.NumberOfDataPoints * varHeader.Count;
    }
    if ( varHeader.ContinuousDataType == 0 ) {
        // values are saved as 16-bit integers
        std::vector<short> values;
        values.resize( totalNumberOfPoints );
        // read the values, 2 bytes per value
        fread( &values[0], totalNumberOfPoints * sizeof( short ), 1, fp );
        // convert values to milliVolts
        for ( size_t i = 0; i < values.size(); ++i ) {
            valuesInMilliVolts.push_back( varHeader.UnitsOffset + ( double )values[i] * varHeader.ADtoUnitsCoefficient );
        }
    } else {
        // values are saved as 32-bit floats
        std::vector<float> values;
        values.resize( totalNumberOfPoints );
        // read the values, 4 bytes per value
        fread( &values[0], totalNumberOfPoints * sizeof( float ), 1, fp );
        // no need to convert values to milliVolts - float values are already in milliVolts
        for ( size_t i = 0; i < values.size(); ++i ) {
            valuesInMilliVolts.push_back( ( double )values[i] );
        }
    }
    return valuesInMilliVolts;
}

void ReadNeuronData( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader )
{
    printf( "Neuron '%s' [%llu timestamps]:", varHeader.Name, varHeader.Count );
    
    // seek to the start of data
    fseek( fp,  varHeader.DataOffset, SEEK_SET );

    std::vector<double> timestampsInSeconds = ReadTimestampsReturnTimestampsInSeconds( fp, varHeader, fileHeader );
    
    for ( size_t i = 0; i < min( 3, timestampsInSeconds.size() ); ++i ) {
        printf( " %.6f", timestampsInSeconds[i] );
    }
    printf( "...\n" );
}

// this is identical to ReadNeuronData
void ReadEventData( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader )
{
    printf( "Event '%s' [%llu timestamps]:", varHeader.Name, varHeader.Count );
    // seek to the start of data
    fseek( fp,  varHeader.DataOffset, SEEK_SET );

    std::vector<double> timestampsInSeconds = ReadTimestampsReturnTimestampsInSeconds( fp, varHeader, fileHeader );

    for ( size_t i = 0; i < min( 3, timestampsInSeconds.size() ); ++i ) {
        printf( " %.6f", timestampsInSeconds[i] );
    }
    printf( "...\n" );
}

void ReadIntervalData( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader )
{
    printf( "Interval var. '%s' [%llu intervals]:", varHeader.Name, varHeader.Count );
    // seek to the start of data
    fseek( fp,  varHeader.DataOffset, SEEK_SET );

    std::vector<double> timestampsOfIntStartsInSeconds = ReadTimestampsReturnTimestampsInSeconds( fp, varHeader, fileHeader );
    std::vector<double> timestampsOfIntEndsInSeconds = ReadTimestampsReturnTimestampsInSeconds( fp, varHeader, fileHeader );

    // print values of the first 3 intervals
    for ( size_t i = 0; i < min( 3, timestampsOfIntStartsInSeconds.size() ); ++i ) {
        printf( " [%.6f,%.6f] ", timestampsOfIntStartsInSeconds[i], timestampsOfIntEndsInSeconds[i] );
    }
    printf( "...\n" );
}

void ReadWaveformData( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader )
{
    printf( "Waveform '%s' [%llu points in waveform, %llu waveforms]:", varHeader.Name, varHeader.NumberOfDataPoints, varHeader.Count );
    // seek to the start of data
    fseek( fp,  varHeader.DataOffset, SEEK_SET );

    std::vector<double> timestampsInSeconds = ReadTimestampsReturnTimestampsInSeconds( fp, varHeader, fileHeader );
    std::vector<double> valuesInMilliVolts = ReadWaveformOrContValuesReturnValuesInMilliVolts( fp, varHeader, fileHeader );

    // print the first 3 values of the first waveform
    for ( size_t i = 0; i < min( 1, timestampsInSeconds.size() ); ++i ) {
        printf( "ts=%.6f;values:", timestampsInSeconds[i] );
        for ( int wavePoint = 0; wavePoint < min( 3, varHeader.NumberOfDataPoints ); ++wavePoint ) {
            printf( "%.4f,", valuesInMilliVolts[ i * varHeader.NumberOfDataPoints + wavePoint] );
        }
    }
    printf( "...\n" );
}

void ReadContinuousData( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader )
{
    printf( "Continuous '%s' [%llu fragments, %llu data points]:", varHeader.Name, varHeader.Count, varHeader.NumberOfDataPoints );
    // seek to the start of data
    fseek( fp,  varHeader.DataOffset, SEEK_SET );

    std::vector<double> fragmentTimestampsInSeconds = ReadTimestampsReturnTimestampsInSeconds( fp, varHeader, fileHeader );

    std::vector<unsigned long long> fragmentIndexes;
    // we have 2 options in saving fragment indexes
    if ( varHeader.ContinuousIndexOfFirstPointInFragmentDataType == 0 ) {
        // indexes are saved as 32-bit unsigned integers
        std::vector<unsigned int> indexes;
        indexes.resize( varHeader.Count );
        // read the values, 4 bytes per value
        fread( &indexes[0], varHeader.Count * sizeof( unsigned int ), 1, fp );
        // convert unsigned long long
        for ( size_t i = 0; i < indexes.size(); ++i ) {
            fragmentIndexes.push_back( indexes[i] );
        }
    } else {
        // indexes are saved as 64-bit unsigned integers
        // this option is not supported by NeuroExplorer yet (as of July 2017, version 5.103)
        fragmentIndexes.resize( varHeader.Count );
        // read the values, 4 bytes per value
        fread( &fragmentIndexes[0], varHeader.Count * sizeof( unsigned long long ), 1, fp );
    }

    std::vector<double> valuesInMilliVolts = ReadWaveformOrContValuesReturnValuesInMilliVolts( fp, varHeader, fileHeader );

    // print timestamps and values of the first 2 data points
    // if the first fragment has more than 1 point
    if ( fragmentTimestampsInSeconds.size() > 0 ) {
        size_t numPointsInFirstFragment = 0;
        if ( fragmentIndexes.size() > 1 ) {
            numPointsInFirstFragment = fragmentIndexes[1];
        } else {
            numPointsInFirstFragment = valuesInMilliVolts.size();
        }
        if ( numPointsInFirstFragment > 1 ) {
            // print values of the first 2 data points
            printf( "ts=%.6f,value=%.3f;", fragmentTimestampsInSeconds[0], valuesInMilliVolts[0] );
            printf( "ts=%.6f,value=%.3f", fragmentTimestampsInSeconds[0] + ( 1.0 / varHeader.SamplingFrequency ), valuesInMilliVolts[0] );
        }
    }
    printf( "...\n" );
}

void ReadMarkerData( FILE* fp, Nex5VarHeader& varHeader, Nex5FileHeader& fileHeader )
{
    printf( "Marker '%s' [%llu values]:", varHeader.Name, varHeader.Count );
    // seek to the start of data
    fseek( fp,  varHeader.DataOffset, SEEK_SET );

    std::vector<double> timestampsInSeconds = ReadTimestampsReturnTimestampsInSeconds( fp, varHeader, fileHeader );

    // reserve vector for field names
    std::vector < std::string > fieldNames;
    fieldNames.resize( varHeader.NumberOfMarkerFields );

    // reserve matrix for field values
    std::vector < std::vector<  std::string > > fieldValuesAsStrings;
    fieldValuesAsStrings.resize( varHeader.NumberOfMarkerFields );
    std::vector < std::vector<  unsigned int > > fieldValuesAsNumbers;
    fieldValuesAsNumbers.resize( varHeader.NumberOfMarkerFields );

    // temp char vectors we will use to read into
    std::vector<char> fieldName;
    std::vector<char> buf;

    // read the marker field names and values
    for ( int field = 0; field < varHeader.NumberOfMarkerFields; field++ ) {
        // read the name of the data field
        fieldName.resize( 65, 0 );
        fread( &fieldName[0], 64, 1, fp );
        fieldNames[field] = ( const char* )&fieldName[0];
        // in .nex5 we have two options for storing marker data
        if ( varHeader.MarkerDataType == 0 ) {
            // markers are stored as strings
            for ( int j = 0; j < varHeader.Count; j++ ) {
                // read the field value for the j-th timestamp
                buf.resize( varHeader.MarkerLength + 1, 0 );
                fread( &buf[0], varHeader.MarkerLength, 1, fp );
                std::string s = ( const char* )&buf[0];
                fieldValuesAsStrings[field].push_back( s );
            }
        } else {
            // markers are store as unsigned 32-bit integers
            fieldValuesAsNumbers[field].resize( varHeader.Count );
            fread( &fieldValuesAsNumbers[field][0], varHeader.Count * sizeof( unsigned int ), 1, fp );
        }
    }
    // print the first marker values
    for ( size_t i = 0; i < min( 1, timestampsInSeconds.size() ); ++i ) {
        printf( "ts=%.6f:", timestampsInSeconds[i] );
        if ( varHeader.MarkerDataType == 0 ) {
            for ( size_t f = 0; f < fieldNames.size(); ++f ) {
                printf( "'%s'='%s',", fieldNames[f].c_str(), fieldValuesAsStrings[f][i].c_str() );
            }
        } else {
            for ( size_t f = 0; f < fieldNames.size(); ++f ) {
                printf( "'%s'=%u,", fieldNames[f].c_str(), fieldValuesAsNumbers[f][i] );
            }
        }
    }
    printf( "...\n" );
}

void PrintHeaderInfo( Nex5FileHeader& fh )
{
    printf( "Nex5 file version: %d\n", fh.Nex5FileVersion );
    printf( "Frequency: %f Hz\n", fh.TimestampFrequency );
    printf( "Number of variables: %d\n", fh.NumberOfVariables );
}

// code that writes various data types to .nex5 file
void WriteNex5File( const char* fname )
{
    FILE* fp = fopen( fname, "wb" );
    if ( fp == 0 ) return;

    // we'll make a .nex5 file with 6 variables:
    // - two spike trains
    // - one interval variable
    // - one continuous variable
    // - one waveform variable
    // - one marker variable

    // first, let's fill the spike trains
    // we will work with time resolution of 100 us or frequency 10000 Hz
    // neuron 1: spikes at 1, 2, and 3 sec
    int spikeTrain1[3] = { 10000, 20000, 30000} ; // in time ticks: 1.*10000, etc.
    // neuron 2: spikes at 5, and 6 sec
    int spikeTrain2[2] = { 50000, 60000} ; // in time ticks

    // interval variable with 2 intervals: [1 sec, 2 sec] and [4 sec, 6 sec]
    int intervalStartTimes[2] = { 10000, 40000 };
    int intervalEndTimes[2]   = { 20000, 60000 };

    // continuous variable: 2 fragments, one starts at 1 sec, another at 7 sec
    int numberOfFragments = 2;
    int fragmentTimestamps[2] = { 10000, 70000 };
    // continuous variable: first fragment has 5 data points, second fragment has 3 data points
    // what we need to store is the index of the first data point in each fragment
    // for the first fragment, the first data point is point number 0,
    // for the second fragment, the first data point is point number 5
    unsigned int indexOfFirstPointInFragment[2] = { 0, 5 };
    // continuous variable: a/d values
    int numberOfDataPointsInContVar = 8;
    short contVarDataPoints[8] = { 0, 1, 2, 3, 4, 0, 4, 6 };

    // waveform variable: 3 waveforms, start at 0.5 sec, 1.5 sec and 2 sec
    int numberOfWaveforms = 3;
    int waveformTimestamps[3] = { 5000, 15000, 20000 };
    // waveform variable: a/d values
    int numberOfDataPointsInWave = 8;
    // waveform values
    short waveform1Values[8] = { 0, 1, 2, 3, 4, 5, 6, 7 };
    short waveform2Values[8] = { 7, 6, 5, 4, 3, 2, 1, 0 };
    short waveform3Values[8] = { 0, -1, -2, -3, -2, -1, 0, 2 };

    // marker variable: 3 markers, each has 2 fields
    // markers are at 1, 2, and 3 sec
    int markerTimestamps[3] = { 10000, 20000, 30000} ; // in time ticks: 1.*10000, etc.
    // names of marker fields;
    char markerFieldNames[2][64];
    memset( markerFieldNames[0], 0, 64 );
    strcpy( markerFieldNames[0], "field1" );
    memset( markerFieldNames[1], 0, 64 );
    strcpy( markerFieldNames[1], "field2" );

    // values of the first marker field:
    char* markerField1Values[] = {"field1value1", "field1value2", "field1value3"};
    // values of the second marker field:
    char* markerField2Values[] = {"field2value1", "field2value2", "field2value3"};

    // first, write file header
    // we will save timestamps as 32-bit integers
    // if you have timestamps that are larger than 2^31, you need to save timestamps
    // as 64-bit integers. To do this, set 
    // Nex5FileHeader.Nex5FileVersion = 502
    // in all variable headers set 
    // Nex5VarHeader.TimestampDataType = 1
    // and set timestampSize = 8
    int timestampSize = 4;

    Nex5FileHeader fileHeader;
    memset( &fileHeader, 0, sizeof( Nex5FileHeader ) );
    char magic[] = "NEX5";
    fileHeader.Nex5MagicNumber = *( int* )magic;

    fileHeader.Nex5FileVersion = 500;
    strcpy( fileHeader.Comment, "test nex5 file" );
    fileHeader.TimestampFrequency = 10000.; // frequency is 10000 Hz, or 100 microseconds time resolution
    fileHeader.RecordingStartTimeInTicks = 0;
    fileHeader.NumberOfVariables = 6;  // we have 6 variables in the file

    fwrite( &fileHeader, sizeof( Nex5FileHeader ), 1, fp );

    // calculate the data offset for the first variable
    // right after all the headers
    int offset = sizeof( Nex5FileHeader ) + fileHeader.NumberOfVariables * sizeof( Nex5VarHeader );

    Nex5VarHeader varheader;
    memset( &varheader, 0, sizeof( Nex5VarHeader ) );
    // if you have timestamps that are larger than 2^31, you need to save timestamps
    // as 64-bit integers. To do this, set 
    // Nex5FileHeader.Nex5FileVersion = 502
    // in all variable headers set 
    // Nex5VarHeader.TimestampDataType = 1

    // neuron 1 header
    varheader.Type = NEX5_VARIABLE_TYPE_NEURON; // neuron type
    varheader.Version = 500;
    strcpy( varheader.Name, "neuron1" ); // variable name
    varheader.DataOffset = offset;
    varheader.Count = 3;  // number of timestamps
    fwrite( &varheader, sizeof( Nex5VarHeader ), 1, fp );

    // adjust offset: add the bytes of the neuron1 timestamp array
    offset += varheader.Count * timestampSize;

    // neuron 2 header
    memset( &varheader, 0, sizeof( Nex5VarHeader ) );
    varheader.Type = NEX5_VARIABLE_TYPE_NEURON; // neuron type
    varheader.Version = 500;
    strcpy( varheader.Name, "neuron2" ); // variable name
    varheader.DataOffset = offset;
    varheader.Count = 2;  // number of timestamps
    fwrite( &varheader, sizeof( Nex5VarHeader ), 1, fp );

    // adjust offset: add the bytes of the neuron2 timestamp array
    offset += varheader.Count * timestampSize;

    // interval header
    memset( &varheader, 0, sizeof( Nex5VarHeader ) );
    varheader.Type = NEX5_VARIABLE_TYPE_INTERVAL; // interval type
    varheader.Version = 500;
    strcpy( varheader.Name, "interval1" ); // variable name
    varheader.DataOffset = offset;
    varheader.Count = 2;  // number of intervals
    fwrite( &varheader, sizeof( Nex5VarHeader ), 1, fp );

    // adjust offset: add the bytes of the interval starts and ends
    offset += varheader.Count * timestampSize * 2;

    // continuous variable header
    memset( &varheader, 0, sizeof( Nex5VarHeader ) );
    varheader.Type = NEX5_VARIABLE_TYPE_CONTINUOUS; // continuous type
    varheader.Version = 500;
    strcpy( varheader.Name, "continuous1" ); // variable name
    varheader.DataOffset = offset;
    varheader.Count = 2;  // number of fragments
    varheader.NumberOfDataPoints = 8; // we have 8 a/d values in 2 fragments
    varheader.SamplingFrequency = 10; // this is a/d frequency, 10 data points per second
    varheader.ADtoUnitsCoefficient = 0.001; // this is a coefficient to convert a/d values to millivolts.
    // 0.001 here means that the stored a/d values are in volts.

    fwrite( &varheader, sizeof( Nex5VarHeader ), 1, fp );

    // adjust offset: add the bytes of fragment timestamps
    offset += varheader.Count * timestampSize;
    // adjust offset: add the bytes of fragment number of values
    offset += varheader.Count * sizeof( int );
    // adjust offset: add the bytes of a/d values
    offset += varheader.NumberOfDataPoints * sizeof( short );

    // waveform variable header
    memset( &varheader, 0, sizeof( Nex5VarHeader ) );
    varheader.Type = NEX5_VARIABLE_TYPE_WAVEFORM; // waveform type
    varheader.Version = 500;
    strcpy( varheader.Name, "wave1" ); // variable name
    varheader.DataOffset = offset;
    varheader.Count = 3;  // number of waveforms
    varheader.NumberOfDataPoints = 8; // we have 8 a/d values in each waveform
    varheader.SamplingFrequency = 1000; // this is a/d frequency, 1000 data points per second
    varheader.ADtoUnitsCoefficient = 1.; // this is a coefficient to convert a/d values to millivolts.
    // 1 here means that the stored a/d values are in millivolts.

    fwrite( &varheader, sizeof( Nex5VarHeader ), 1, fp );

    // adjust offset: add the bytes of waveform timestamps
    offset += varheader.Count * timestampSize;
    // adjust offset: add the bytes of a/d values
    offset += varheader.Count * varheader.NumberOfDataPoints * sizeof( short );

    // marker header
    memset( &varheader, 0, sizeof( Nex5VarHeader ) );
    varheader.Type = NEX5_VARIABLE_TYPE_MARKER; // marker type
    varheader.Version = 500;
    strcpy( varheader.Name, "marker1" ); // variable name
    varheader.DataOffset = offset;
    varheader.Count = 3;  // number of markers (timestamps)
    varheader.NumberOfMarkerFields = 2; // number of marker fields for each timestamp
    varheader.MarkerLength = 16; // number of bytes for each field string
    fwrite( &varheader, sizeof( Nex5VarHeader ), 1, fp );

    // adjust offset: add the bytes of timestamps
    offset += varheader.Count * timestampSize;
    // adjust offset: add the bytes of field names (64 bytes per field) and field values
    offset += varheader.NumberOfMarkerFields * ( 64 + varheader.Count * varheader.MarkerLength );


    // now, write the data in the same order as we wrote the headers
    // first spike train
    fwrite( spikeTrain1, 3 * timestampSize, 1, fp );
    // second spike train
    fwrite( spikeTrain2, 2 * timestampSize, 1, fp );

    // interval variable -- for the intervals, first we write starts, then ends
    fwrite( intervalStartTimes, 2 * timestampSize, 1, fp );
    fwrite( intervalEndTimes, 2 * timestampSize, 1, fp );

    // for the continuous variable, first we write fragment timestamps,
    // then fragment counts, and finally, a/d values
    fwrite( fragmentTimestamps, numberOfFragments * timestampSize, 1, fp );
    fwrite( indexOfFirstPointInFragment, numberOfFragments * sizeof( unsigned int ), 1, fp );
    fwrite( contVarDataPoints, numberOfDataPointsInContVar * sizeof( short ), 1, fp );

    // for waveform variable, first we write w/f timestamps,
    // then, a/d values
    fwrite( waveformTimestamps, numberOfWaveforms * timestampSize, 1, fp );
    fwrite( waveform1Values, 8 * sizeof( short ), 1, fp );
    fwrite( waveform2Values, 8 * sizeof( short ), 1, fp );
    fwrite( waveform3Values, 8 * sizeof( short ), 1, fp );

    // for the marker variable, first we write the timestamps,
    fwrite( markerTimestamps, 3 * timestampSize, 1, fp );
    // then, for each field --
    // field name (we have to use exactly 64 bytes for each name)
    fwrite( markerFieldNames[0], 64, 1, fp );
    // then, marker field values for the first field
    // we have to use exactly varheader.MarkerLength bytes for each value
    char buf[16];
    for ( int i = 0; i < 3; i++ ) {
        strcpy( buf, markerField1Values[i] );
        fwrite( buf, 16, 1, fp );
    }
    // field name for the second filed
    fwrite( markerFieldNames[1], 64, 1, fp );
    // finally, marker field values for the second field
    // we have to use exactly varheader.MarkerLength bytes for each value
    for ( int i = 0; i < 3; i++ ) {
        strcpy( buf, markerField2Values[i] );
        fwrite( buf, 16, 1, fp );
    }

    // after all the binary data is written, we can add an optional string with metadata
    // store the start of the location of metadata
    fileHeader.MetadataOffset = ftell( fp );

    // write metadata
    // Here is an example of metadata describing a data file and two neuron variables ('neuron1' and 'neuron2')
    //    with probe position, wire number and unit number specified:
    // see HowToReadAndWriteNex5Files.txt for the description of metadata

    // to avoid using \" for all the required double quotes (which will make the metadata string in C format unreadable in this file)
    // we use single quotes and then replace them with double quotes
    std::string metaData = "{'file':{'recordingStartTime':'2015-07-0214:49:02.000'},'variables':[{'name':'neuron1','probe':{'position':{'x':12.5,'y':12.5},'wireNumber':1},'unitNumber':2},{'name':'neuron2','probe':{'position':{'x':37.5,'y':12.5},'wireNumber':3},'unitNumber':4 }]}";

    std::replace( metaData.begin(), metaData.end(), '\'', '\"' );

    // write the metadata
    fwrite( metaData.c_str(), metaData.length(), 1, fp );

    // go the beginning of the file and overwrite fileHeader with the information on the location of metadata
    fseek( fp,  0, SEEK_SET );
    fwrite( &fileHeader, sizeof( Nex5FileHeader ), 1, fp );

    printf( "created .nex5 file '%s'\n", fname );
    fclose( fp );
}
